/*
 * Copyright Camunda Services GmbH and/or licensed to Camunda Services GmbH under
 * one or more contributor license agreements. See the NOTICE file distributed
 * with this work for additional information regarding copyright ownership.
 * Licensed under the Zeebe Community License 1.0. You may not use this file
 * except in compliance with the Zeebe Community License 1.0.
 */
package io.zeebe.engine.processing.bpmn;

import io.zeebe.engine.Loggers;
import io.zeebe.engine.processing.bpmn.behavior.BpmnStateBehavior;
import io.zeebe.engine.state.instance.ElementInstance;
import io.zeebe.protocol.record.intent.WorkflowInstanceIntent;
import io.zeebe.protocol.record.value.BpmnElementType;
import io.zeebe.util.Either;
import org.slf4j.Logger;

/**
 * A check to prevent concurrent state transitions of a workflow instance.
 *
 * <p>A workflow instance can be have concurrent state transitions if a user command is received
 * (e.g. cancel workflow instance) or if an internal/external event is triggered (e.g. timer
 * boundary event). In this case, the current workflow instance processing needs to be interrupted
 * be avoid an inconsistent state.
 */
public final class WorkflowInstanceStateTransitionGuard {

  private static final Logger LOGGER = Loggers.WORKFLOW_PROCESSOR_LOGGER;

  private final BpmnStateBehavior stateBehavior;

  public WorkflowInstanceStateTransitionGuard(final BpmnStateBehavior stateBehavior) {
    this.stateBehavior = stateBehavior;
  }

  /**
   * Checks if a workflow instance event can be processed based on the current state.
   *
   * @return {@code true} if the transition is valid.
   */
  public boolean isValidStateTransition(final BpmnElementContext context) {
    final var result = checkStateTransition(context);

    // log the reason for better debugging
    result.ifLeft(
        violation ->
            LOGGER.debug(
                "Don't process event because of an illegal state transition: {} [context: {}]",
                violation,
                context));

    return result.isRight();
  }

  private Either<String, ?> checkStateTransition(final BpmnElementContext context) {
    switch (context.getIntent()) {
      case ELEMENT_ACTIVATING:
      case ELEMENT_ACTIVATED:
      case ELEMENT_COMPLETING:
      case ELEMENT_COMPLETED:
        return hasElementInstanceWithState(context, context.getIntent())
            .flatMap(ok -> hasActiveFlowScopeInstance(context));

      case ELEMENT_TERMINATING:
      case ELEMENT_TERMINATED:
        return hasElementInstanceWithState(context, context.getIntent());

      case EVENT_OCCURRED:
        if (context.getBpmnElementType() == BpmnElementType.START_EVENT) {
          return Either.right(null);
        }
        return hasElementInstanceWithState(context, WorkflowInstanceIntent.ELEMENT_ACTIVATED);

      case SEQUENCE_FLOW_TAKEN:
        return hasActiveFlowScopeInstance(context);

      default:
        return Either.left(
            String.format(
                "Expected event to have a workflow instance intent but was '%s'",
                context.getIntent()));
    }
  }

  private Either<String, ElementInstance> getElementInstance(final BpmnElementContext context) {
    final var elementInstance = stateBehavior.getElementInstance(context);
    if (elementInstance != null) {
      return Either.right(elementInstance);

    } else {
      return Either.left(
          String.format(
              "Expected element instance with key '%d' to be present in state but not found.",
              context.getElementInstanceKey()));
    }
  }

  private Either<String, ElementInstance> getFlowScopeInstance(final BpmnElementContext context) {
    final var flowScopeInstance = stateBehavior.getFlowScopeInstance(context);
    if (flowScopeInstance != null) {
      return Either.right(flowScopeInstance);

    } else {
      return Either.left(
          String.format(
              "Expected flow scope instance with key '%d' to be present in state but not found.",
              context.getElementInstanceKey()));
    }
  }

  private Either<String, ElementInstance> hasElementInstanceInState(
      final ElementInstance elementInstance, final WorkflowInstanceIntent expectedState) {
    final var currentState = elementInstance.getState();
    if (currentState != expectedState) {
      return Either.left(
          String.format(
              "Expected element instance to be in state '%s' but was '%s'.",
              expectedState, currentState));

    } else {
      return Either.right(elementInstance);
    }
  }

  private Either<String, ElementInstance> hasFlowScopeInstanceInState(
      final ElementInstance flowScopeInstance, final WorkflowInstanceIntent expectedState) {
    final var currentState = flowScopeInstance.getState();
    if (currentState != expectedState) {
      return Either.left(
          String.format(
              "Expected flow scope instance to be in state '%s' but was '%s'.",
              expectedState, currentState));

    } else {
      return Either.right(flowScopeInstance);
    }
  }

  private Either<String, ElementInstance> hasNonInterruptedFlowScope(
      final ElementInstance flowScopeInstance, final BpmnElementContext context) {
    final var interruptingEventKey = flowScopeInstance.getInterruptingEventKey();
    if (interruptingEventKey > 0 && interruptingEventKey != context.getElementInstanceKey()) {
      return Either.left(
          String.format(
              "Expected flow scope instance to be not interrupted but was interrupted by an event with key '%d'.",
              interruptingEventKey));

    } else {
      return Either.right(flowScopeInstance);
    }
  }

  private Either<String, ?> hasElementInstanceWithState(
      final BpmnElementContext context, final WorkflowInstanceIntent expectedState) {
    // a shortcut to improve readability
    return getElementInstance(context)
        .flatMap(elementInstance -> hasElementInstanceInState(elementInstance, expectedState));
  }

  private Either<String, ?> hasActiveFlowScopeInstance(final BpmnElementContext context) {
    // a shortcut to improve readability
    if (context.getBpmnElementType() == BpmnElementType.PROCESS) {
      // a process has no flow scope instance
      return Either.right(null);

    } else {
      return getFlowScopeInstance(context)
          .flatMap(
              flowScopeInstance ->
                  hasFlowScopeInstanceInState(
                      flowScopeInstance, WorkflowInstanceIntent.ELEMENT_ACTIVATED))
          .flatMap(flowScopeInstance -> hasNonInterruptedFlowScope(flowScopeInstance, context));
    }
  }

  public void registerStateTransition(
      final BpmnElementContext context, final WorkflowInstanceIntent newState) {
    switch (newState) {
      case ELEMENT_ACTIVATING:
      case ELEMENT_ACTIVATED:
      case ELEMENT_COMPLETING:
      case ELEMENT_COMPLETED:
      case ELEMENT_TERMINATING:
      case ELEMENT_TERMINATED:
        updateElementInstanceState(context, newState);
        break;

      default:
        // other transitions doesn't change the state of an element instance
        break;
    }
  }

  private void updateElementInstanceState(
      final BpmnElementContext context, final WorkflowInstanceIntent newState) {

    stateBehavior.updateElementInstance(
        context, elementInstance -> elementInstance.setState(newState));
  }
}
